const standardOptions = {
    method: "POST",
    headers: {
        "Content-type": "application/json",
        Accept: "application/json"
    }
};
const randChannelName = () => Math.ceil(Math.random() * 1000000).toString();
export const noOpTag = (strings, ...exps) => {
    return strings.map((str, i) => str + (i in exps ? exps[i] : "")).join("");
};
export const gql = noOpTag;
export class Api {
    constructor(url, apiOptions = {}) {
        const options = typeof apiOptions === "string"
            ? { wsUrl: apiOptions }
            : apiOptions;
        if (!options.fetch && typeof window === "undefined") {
            throw new Error("No fetch implementation provided and not operating in browser context");
        }
        options.fetch = options.fetch || window.fetch.bind(window);
        const protocol = url.match(/^(https?):\/\//)[1];
        if (!protocol)
            throw new Error(`Unexpected API URL [${url}]`);
        const isSecure = protocol.match(/s$/);
        this.url = url;
        this.wsUrl = options.wsUrl || [`ws${isSecure ? "s" : ""}:`, url.split("//").slice(1)].join("//");
        this.fetch = options.fetch;
        this.headers = options.headers;
        this.onError = (err) => {
            const msg = typeof err === "string"
                ? err
                : (err?.message) || "GraphQL error. Try: Check network connection / Turn off ad blockers";
            if (typeof options.onError === "function") {
                options.onError(msg, err);
            }
            err = err instanceof Error ? err : new Error(msg);
            throw err;
        };
    }
    get log() {
        return this.socket.log;
    }
    set log(fn) {
        if (typeof fn === "function") {
            if (this.socket)
                this.socket.log = fn;
        }
    }
    async run(query, variables = {}) {
        const headers = { ...standardOptions.headers, ...this.headers };
        let response;
        try {
            response = await this.fetch(this.url, {
                ...standardOptions,
                headers,
                body: JSON.stringify({ query, variables })
            });
        }
        catch (err) {
            this.onError(err);
        }
        const { errors, data } = await response.json();
        if (Array.isArray(errors) && errors.length) {
            this.onError(errors[0]);
        }
        return data;
    }
    subscribe(query, variables = {}, channelName = randChannelName()) {
        const message = {
            id: channelName,
            type: "start",
            payload: {
                query,
                variables
            }
        };
        const startSub = () => this.socket.webSocket.send(JSON.stringify(message));
        if (this.socket) {
            if (this.socket.subscriptions[channelName]) {
                throw new Error(`Subscription already exists for channel [${channelName}]`);
            }
            if (this.socket.connected) {
                startSub();
            }
            else {
                this.socket.connectedHandlers.push(startSub);
            }
        }
        else {
            this.socket = new Socket(this.wsUrl, this.headers, startSub);
            if (this.log)
                this.socket.log = this.log;
        }
        const socketSubscription = this.socket.subscriptions[channelName] = {};
        const result = {
            onData(handler) {
                socketSubscription.dataHandler = handler;
                return result;
            },
            close() {
                delete this.socket.subscriptions[channelName];
            }
        };
        return result;
    }
}
class Socket {
    constructor(wsUrl, headers = {}, onConnected = () => { }) {
        this.connected = false;
        this.connectedHandlers = onConnected ? [onConnected] : [];
        this.subscriptions = {};
        this.webSocket = new WebSocket(wsUrl, "graphql-subscriptions");
        this.webSocket.onopen = () => {
            const message = {
                type: "connection_init",
                payload: headers
            };
            this.webSocket.send(JSON.stringify((message)));
        };
        this.webSocket.onmessage = event => {
            const data = JSON.parse(event.data);
            let msg;
            if (["subscription_fail", "error"].includes(data.type)) {
                const msg = `GraphQL ${data.type} for channel ${data.id} ${(data?.payload.message) || ""}`;
                throw new Error(msg);
            }
            else if (data.type === "data") {
                if (this.subscriptions[data.id] && typeof this.subscriptions[data.id].dataHandler === "function") {
                    return this.subscriptions[data.id].dataHandler(data.payload.data);
                }
                else
                    msg = `data received for channel [${data.id}] with no subscription handler`;
            }
            else if (data.type !== "ka") {
                const messages = {
                    connection_ack: `[${data.id}] connection_ack, the handshake is complete`,
                    init_fail: `[${data.id}] init_fail returned from the WebSocket server`,
                    subscription_success: `[${data.id}] subscription_success`
                };
                msg = messages[data.type];
                msg = msg || `unexpected message type [${data.type}] received from WebSocket server`;
                if (data.type === "connection_ack" && typeof onConnected === "function") {
                    this.connected = true;
                    this.connectedHandlers.forEach(connHandler => connHandler());
                }
            }
            if (msg && typeof this.log === "function")
                this.log(`graphqlx: ${msg}`);
        };
    }
}
//# sourceMappingURL=graphqlex.js.map